import tempfile
import time
import webbrowser
from pathlib import Path
from typing import Union


class HTMLFigure:
    def __init__(self, html: str):
        self.html = html

    def show(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = Path(temp_dir).joinpath("fig.html")
            self.write_html(file_name)
            webbrowser.open("file://" + str(file_name.absolute()), new=2)
            time.sleep(2)

    def write_html(self, path: Union[str, Path]):
        path = Path(path)
        with path.open("w") as out_file:
            out_file.write(self.html)

    def __repr_html__(self):
        return self.html



from functools import cache
from typing import Callable

import numpy as np
from scipy import optimize
from scipy.sparse import issparse


def decomposition_gaussian_bic(
    n_components: int, decomp_class, X, random_state: int = 42
) -> float:
    """Computes Bayesian information criterion for
    a decomposition model with assumed Gaussian noise.

    Parameters
    ----------
    n_components: int
        Number of components in the model.
    decomp_class
        Class of the decomposition model, e.g. `decomp_class=FastICA`.
    X: array
        Training data to decompose.
    random_state: int, default 42
        Seed to pass to the decomposition algorithm.
    """
    decomp = decomp_class(n_components=n_components, random_state=42)
    doc_topic = decomp.fit_transform(X)
    if not hasattr(decomp, "reconstruction_err_"):
        X_reconstruction = decomp.inverse_transform(doc_topic)
        # Computing residual sum of squares
        if issparse(X):
            X = np.asarray(X.todense())
        rss = np.sum(np.square(X - X_reconstruction))
    else:
        rss = np.square(decomp.reconstruction_err_)
    n_params = decomp.components_.size + np.prod(doc_topic.shape)
    n_datapoints = np.prod(X.shape)
    # Computing Bayesian information criterion assuming IID
    bic = n_params * np.log(n_datapoints) + n_datapoints * np.log(
        rss / n_datapoints
    )
    return bic


def optimize_n_components(
    f_ic: Callable[int, float], min_n: int = 2, max_n: int = 250, verbose=False
) -> int:
    """Optimizes the nuber of components using the Brent minimum finding
    algorithm given an information criterion.
    The bracket is found by exponentially incrementing n_components.

    Parameters
    ----------
    f_ic: Callable[int, float]
        Information criterion given N components.
    min_n: int
        Minimum value of N.
    max_n: int
        Maximum value of N, in case the algorithm doesn't converge.
    """
    if verbose:
        print("Optimizing N based on an information criterion...")

    # Caching and adding debugging statements
    @cache
    def _f_ic(n_components) -> float:
        # Making sure n_components is an integer
        # The optimization algorithm definitely give it a float
        n_components = int(n_components)
        val = f_ic(n_components)
        if verbose:
            print(f" - IC(N={n_components})={val:.2f}")
        return val

    # Finding bracket
    if verbose:
        print(" - Finding bracket...")
    low = min_n
    n_comp = 2
    while not _f_ic(n_comp) < _f_ic(min_n):
        n_comp += 1
        if n_comp >= 10:
            if verbose:
                print(
                    " - Couldn't find lower value than n=1 up to n=10, stopping."
                )
            return 1
    middle = n_comp
    current = _f_ic(middle)
    inc = 5
    while not current > _f_ic(middle):
        n_comp += inc
        if n_comp >= max_n:
            if verbose:
                print(f" - Bracket didn't converge, returning max N: {max_n}")
            return max_n
        current = _f_ic(n_comp)
        if current < _f_ic(middle):
            low = n_comp - inc
            middle = n_comp
        inc *= 2
    bracket = low, middle, n_comp
    if verbose:
        print(f" - Running optimization with bracket: {bracket}")
    # Optimizing
    res = optimize.minimize_scalar(
        _f_ic,
        method="brent",
        bracket=(low, middle, n_comp),
        options=dict(xtol=0.2),
    )
    if verbose:
        print(f" - Converged after {res.nit} iterations.")
    return int(res.x)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from rich.console import Console
from rich.progress import track

NAMER_PROMPT = """
You will be tasked with naming a topic.
Based on the provided information, create a short label that best summarizes the topics.
Only respond with a short, human readable topic name and nothing else.

The topic is described by the following set of keywords: {keywords}.
"""

DESCRIPTION_PROMPT = """
Based on the provided information, summarize the content of the topic in a couple of sentences.
Do NOT name the topic, just describe its contents.

The topic is described by the following set of keywords: {keywords}.
"""

DOCUMENT_TEMPLATE = """
In addition the topic is characterized by the following documents:
{documents}
"""

CONTEXT_TEMPLATE = """
Additional context for the task provided by the user: {context}
"""

SUMMARY_PROMPT = """
Summarize the following document: {document}
"""

DEFAULT_SYSTEM_PROMPT = """
You are a topic analyzer.
When the user asks you to name a topic, you respond with a name for the topic they describe.
When they ask you to describe the topic briefly, you respond with a couple of sentences explaining what the topic is about.
When they ask you to summarize a document, you respond with a brief summary.
"""


@dataclass
class AnalysisResults:
    """Container class for results of topic analysis.

    Attributes
    ----------
    topic_names: list[str]
        Generated topic names.
    topic_descriptions: list[str]
        Genreated topic descriptions.
    document_summaries: list[list[str]], default None
        Summaries of top 10 documents for each topic, when use_summaries is enabled.
    """

    topic_names: list[str]
    topic_descriptions: list[str]
    document_summaries: Optional[list[list[str]]] = None

    def to_dict(self) -> dict:
        """Returns the analysis result as a dictionary"""
        res = dict(
            topic_names=self.topic_names,
            topic_descriptions=self.topic_descriptions,
        )
        if self.document_summaries is not None:
            res["document_summaries"] = self.document_summaries
        return res

    def to_df(self):
        """Turns analysis result object into a dataframe"""
        try:
            import pandas as pd
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "You need to pip install pandas to be able to use dataframes."
            )
        return pd.DataFrame(self.to_dict())


class Analyzer(ABC):
    system_prompt = DEFAULT_SYSTEM_PROMPT
    summary_prompt = SUMMARY_PROMPT
    namer_prompt = NAMER_PROMPT
    description_prompt = DESCRIPTION_PROMPT
    context = None
    use_summaries = False

    @abstractmethod
    def generate_text(self, prompt: str) -> str:
        """Generates response to a given prompt."""
        pass

    def summarize_document(self, document: str) -> str:
        """Summarizes document so that analysis becomes easier."""
        prompt = self.summary_prompt.format(document=document)
        return self.generate_text(prompt)

    def describe_topic(
        self,
        keywords: list[str],
        documents: Optional[list] = None,
    ):
        """Gives abstract summarization of topic content."""
        _keys = ", ".join(keywords)
        prompt = self.description_prompt.format(keywords=_keys)
        if documents:
            prompt += self.template_documents(documents)
        if self.context:
            prompt += CONTEXT_TEMPLATE.format(context=self.context)
        return self.generate_text(prompt)

    def name_topic(
        self,
        keywords: list[str],
        documents: Optional[list] = None,
    ) -> str:
        """Names one topic based on top descriptive aspects."""
        _keys = ", ".join(keywords)
        prompt = self.namer_prompt.format(keywords=_keys)
        if documents:
            prompt += self.template_documents(documents)
        if self.context:
            prompt += CONTEXT_TEMPLATE.format(context=self.context)
        return self.generate_text(prompt)

    def name_topics(
        self,
        keywords: list[list[str]],
        documents: list[list[str]] = None,
    ) -> list[str]:
        """Names all topics based on top descriptive terms.

        Parameters
        ----------
        keywords: list[list[str]]
            Top K highest ranking terms on the topics.
        documents: list[list[str]], optional
            Top K relevant documents to each topic.

        Returns
        -------
        list[str]
            Topic names returned by the namer.
        """
        names = []
        if documents is not None:
            key_doc = list(zip(keywords, documents))
            for keys, docs in track(key_doc, description="Naming topics..."):
                names.append(self.name_topic(keys, documents=docs))
        else:
            for keys in track(keywords, description="Naming topics..."):
                names.append(self.name_topic(keys))
        return names

    def template_documents(self, documents: list[str]) -> str:
        doc_list = "\n".join([f" - {doc}" for doc in documents])
        return """
        In addition the topic is characterized by the following documents:
        {documents}
        """.format(
            documents=doc_list
        )

    def analyze_topics(
        self,
        keywords: list[list[str]],
        documents: Optional[list[list[str]]] = None,
        use_summaries: Optional[bool] = None,
    ) -> AnalysisResults:
        """Analyzes topic model with a language model.
        Generates topic names, descriptions and document summaries (optional).

        Parameters
        ----------
        keywords: list[list[str]]
            Keywords for each topic.
        documents: list[list[str]], optional
            Top documents for each topic.
        use_summaries: bool, optional
            Indicates whether the analyzer should summarize documents
            prior to analyzing the topic.

        Returns
        -------
        dict
            Dictionary containing `topic_names`, `topic_descriptions` and `document_summaries` if relevant.
        """
        console = Console()
        output = {"topic_names": [], "topic_descriptions": []}
        use_summaries = (
            use_summaries if use_summaries is not None else self.use_summaries
        )
        if documents is not None:
            if use_summaries:
                output["document_summaries"] = []
                for docs in track(
                    documents, description="Summarizing documents"
                ):
                    _sums = []
                    for doc in docs:
                        _sums.append(self.summarize_document(doc))
                    output["document_summaries"].append(_sums)
                console.log("Documents summarized.")
                # Updating parameter so summaries are used down-stream
                documents = output["document_summaries"]
            # Organizing into a list so we can iterate and know the length at the same time.
            key_doc_pairs = list(zip(keywords, documents))
            for keys, docs in track(
                key_doc_pairs, description="Generating topic names"
            ):
                output["topic_names"].append(
                    self.name_topic(keys, documents=docs)
                )
            console.log("Topic names generated.")
            for keys, docs in track(
                key_doc_pairs, description="Generating topic descriptions."
            ):
                output["topic_descriptions"].append(
                    self.describe_topic(keys, documents=docs)
                )
            console.log("Topic descriptions generated.")
        else:
            for keys in track(keywords, description="Naming"):
                output["topic_names"].append(
                    self.name_topic(keys, documents=None)
                )
            console.log("Topic names generated.")
            for keys in track(keywords, description="Describing topics."):
                output["topic_descriptions"].append(
                    self.describe_topic(keys, documents=None)
                )
            console.log("Topic descriptions generated.")
        return AnalysisResults(**output)



import json
import tempfile
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, List, Optional, Union

import joblib
import numpy as np
from huggingface_hub import HfApi
from sentence_transformers import SentenceTransformer
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError

from turftopic.container import TopicContainer
from turftopic.data import TopicData
from turftopic.encoders import ExternalEncoder
from turftopic.serialization import create_readme, get_package_versions

Encoder = Union[ExternalEncoder, SentenceTransformer]


class ContextualModel(BaseEstimator, TransformerMixin, TopicContainer):
    """Base class for contextual topic models in Turftopic."""

    @property
    def has_negative_side(self) -> bool:
        return False

    def encode_documents(self, raw_documents: Iterable[str]) -> np.ndarray:
        """Encodes documents with the sentence encoder of the topic model.

        Parameters
        ----------
        raw_documents: iterable of str
            Textual documents to encode.

        Return
        ------
        ndarray of shape (n_documents, n_dimensions)
            Matrix of document embeddings.
        """
        if not hasattr(self.encoder_, "encode"):
            return self.encoder.get_text_embeddings(list(raw_documents))
        return self.encoder_.encode(raw_documents)

    @abstractmethod
    def fit_transform(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Fits model and infers topic importances for each document.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        y: None
            Ignored, exists for sklearn compatibility.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.

        Returns
        -------
        ndarray of shape (n_documents, n_topics)
            Document-topic matrix.
        """
        pass

    def fit(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ):
        """Fits model on the given corpus.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        y: None
            Ignored, exists for sklearn compatibility.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.
        """
        self.fit_transform(raw_documents, y, embeddings)
        return self

    def get_vocab(self) -> np.ndarray:
        """Get vocabulary of the model.

        Returns
        -------
        ndarray of shape (n_vocab)
            All terms in the vocabulary.
        """
        return self.vectorizer.get_feature_names_out()

    def get_feature_names_out(self) -> np.ndarray:
        """Get topic ids.

        Returns
        -------
        ndarray of shape (n_topics)
            IDs for each output feature of the model.
            This is useful, since some models have outlier
            detection, and this gets -1 as ID, instead of
            its index.
        """
        n_topics = self.components_.shape[0]
        try:
            classes = self.classes_
        except AttributeError:
            classes = list(range(n_topics))
        return np.asarray(classes)

    def prepare_topic_data(
        self,
        corpus: List[str],
        embeddings: Optional[np.ndarray] = None,
    ) -> TopicData:
        """Produces topic inference data for a given corpus, that can be then used and reused.
        Exists to allow visualizations out of the box with topicwizard.

        Parameters
        ----------
        corpus: list of str
            Documents to infer topical content for.
        embeddings: ndarray of shape (n_documents, n_dimensions)
            Embeddings of documents.

        Returns
        -------
        TopicData
            Information about topical inference in a dictionary.
        """
        if embeddings is None:
            embeddings = self.encode_documents(corpus)
        try:
            document_topic_matrix = self.transform(
                corpus, embeddings=embeddings
            )
        except (AttributeError, NotFittedError):
            document_topic_matrix = self.fit_transform(
                corpus, embeddings=embeddings
            )
        dtm = self.vectorizer.transform(corpus)  # type: ignore
        try:
            classes = self.classes_
        except AttributeError:
            classes = list(range(self.components_.shape[0]))
        res = TopicData(
            corpus=corpus,
            document_term_matrix=dtm,
            vocab=self.get_vocab(),
            document_topic_matrix=document_topic_matrix,
            document_representation=embeddings,
            topic_term_matrix=self.components_,  # type: ignore
            transform=getattr(self, "transform", None),
            topic_names=self.topic_names,
            classes=classes,
            has_negative_side=self.has_negative_side,
            hierarchy=getattr(self, "hierarchy", None),
        )
        return res

    def to_disk(self, out_dir: Union[Path, str]):
        """Persists model to directory on your machine.

        Parameters
        ----------
        out_dir: Path | str
            Directory to save the model to.
        """
        out_dir = Path(out_dir)
        out_dir.mkdir(exist_ok=True)
        package_versions = get_package_versions()
        with out_dir.joinpath("package_versions.json").open("w") as ver_file:
            ver_file.write(json.dumps(package_versions))
        joblib.dump(self, out_dir.joinpath("model.joblib"))

    def push_to_hub(self, repo_id: str):
        """Uploads model to HuggingFace Hub

        Parameters
        ----------
        repo_id: str
            Repository to upload the model to.
        """
        api = HfApi()
        api.create_repo(repo_id, exist_ok=True)
        with tempfile.TemporaryDirectory() as tmp_dir:
            readme_path = Path(tmp_dir).joinpath("README.md")
            with readme_path.open("w") as readme_file:
                readme_file.write(create_readme(self, repo_id))
            self.to_disk(tmp_dir)
            api.upload_folder(
                folder_path=tmp_dir,
                repo_id=repo_id,
                repo_type="model",
            )



import re
import tempfile
import time
import webbrowser
from pathlib import Path
from typing import Optional

import numpy as np
from sklearn.preprocessing import scale

from turftopic.utils import sanitize_for_html

CUSTOM_CSS = """
.row {
    display : flex;
    align-items : center;
}
.box {
    height:10px;
    width:10px;
    border-radius:2px;
    margin-right:7px;
    margin-top:5px;
    margin-bottom:5px;
    padding:0px 0 1px 0;
    text-align:center;
    color: white;
    font-size: 14px;
    cursor: pointer;
}
.description {
    position: absolute;
    bottom: 0;
    left: 0;
    text-align: justify;
    padding: 20px;
    margin: 15px;
    max-width: 40%;
    background-color: #ffffff;
    border-radius:15px;
    box-shadow: 0px 0px 5px 5px rgba(0,0,0,0.05);
}
h3 {
    margin: 0px;
    margin-bottom: 4px;
    padding: 0px;
}
p {
    margin: 0px;
    padding: 0px;
}
#legend {
        position: absolute;
        top: 0;
        right: 0;
    }
#title-container {
    max-width: 75%;
}
.topic-tree-header {
    display: none;
}
.topic-tree-close-btn {
    display: none;
}
.bullet {
    padding: 0px;
}
#topic-tree-body > .nested {
    display: block;
    padding: 0px;
    margin-right: 10px;
}
#topic-tree-body > li {
    margin: 0px;
}
#topic-tree {
    padding: 5px;
}
#topic-tree-body {
    padding: 0px;
}
#topic-keywords {
    font-style: italic;
    margin-bottom: 2px;
    text-align: left;
}
#topic-percent {
    margin-top: 10px;
    margin-right: 10px;
}
#percent-message {
    margin-top: 10px;
}
"""


def _labels_to_indices(labels, classes):
    n_classes = len(classes)
    class_to_index = dict(zip(classes, np.arange(n_classes)))
    return np.array([class_to_index[label] for label in labels])


def build_datamapplot(
    coordinates: np.ndarray,
    topic_names: list[str],
    labels: np.ndarray,
    classes: np.ndarray,
    top_words: Optional[list[list[str]]] = None,
    topic_descriptions: Optional[list[str]] = None,
    font_family: str = "Merriweather",
    enable_topic_tree=True,
    topic_tree_kwds={
        "color_bullets": True,
    },
    cluster_boundary_polygons=False,
    cluster_boundary_line_width=6,
    polygon_alpha=2,
    **kwargs,
):
    """Builds a Turftopic interactive datamapplot.

    Parameters
    ----------
    coordinates: np.ndarray
        X and Y coordinates of datapoints.
    topic_names: list[str]
        Names of topics in the model.
    labels: np.ndarray
        Topic labels for each datapoint (topic_id for each point)
    classes: np.ndarray
        List of topic IDs in the model.
    top_words: list[list[str]], optional
        List of top keywords for each topic.
    topic_descriptions: list[str], optional
        List of descriptions for the given topics.
    """
    try:
        import datamapplot
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "You need to install datamapplot to be able to use plot_clusters_datamapplot()."
        ) from e
    coordinates = scale(coordinates) * 4
    indices = _labels_to_indices(labels, classes)
    labels = np.array(topic_names)[indices]
    if -1 in classes:
        i_outlier = np.where(classes == -1)[0][0]
        kwargs["noise_label"] = topic_names[i_outlier]
    # Calculating how much of the corpus is made up of a topic
    percentages = []
    for label in topic_names:
        percentages.append(100 * np.sum(labels == label) / len(labels))
    # Sanitizing the names so they don't mess up the HTML
    topic_names = [sanitize_for_html(name) for name in topic_names]
    custom_js = ""
    custom_js += "const nameToPercent = new Map();\n"
    for name, percent in zip(topic_names, percentages):
        custom_js += 'nameToPercent.set("{name}", {percent});\n'.format(
            name=name,
            percent=percent,
        )
    custom_js += "const nameToDesc = new Map();\n"
    if topic_descriptions is not None:
        topic_descriptions = [
            sanitize_for_html(desc) for desc in topic_descriptions
        ]
        for topic_id, name, desc in zip(
            classes, topic_names, topic_descriptions
        ):
            custom_js += 'nameToDesc.set("{name}", "{desc}");\n'.format(
                name=name,
                desc=desc,
            )
    custom_js += "const nameToKeywords = new Map();\n"
    if top_words is not None:
        for name, words in zip(topic_names, top_words):
            custom_js += (
                'nameToKeywords.set("{name}", "{keywords}");\n'.format(
                    name=name, keywords=", ".join(words)
                )
            )
    custom_html = ""
    custom_html += """
<div class="description">
    <h3 id="topic-name">{topic_name}</h3>
    <p id="topic-keywords">{topic_keywords}</p>
    <progress id="topic-percent" value="{percentage:.2f}" max="100"></progress>
    <b id="percent-message">{percentage:.2f}% of all documents</b>
    <hr>
    <p id="topic-description">{topic_description}</p>
</div>
    """.format(
        topic_name=topic_names[1],
        topic_description=(
            topic_descriptions[1] if topic_descriptions is not None else ""
        ),
        topic_keywords=(
            "Keywords: "
            + (", ".join(top_words[1]) if top_words is not None else "")
        ),
        percentage=percentages[1],
    )
    custom_js += """
setTimeout(function(){
    const labelNodes = Array.from(document.getElementsByClassName('topic-tree-label'))
    labelNodes.forEach(button => {
        console.log("Found this button")
        button.addEventListener('click', function(event) {
            const topicName = document.getElementById("topic-name");
            const topicDesc = document.getElementById("topic-description");
            const topicKeywords = document.getElementById("topic-keywords");
            const topicPercent = document.getElementById("topic-percent");
            const percentMessage = document.getElementById("percent-message");
            const name = button.textContent.replace(/[\\n\\r\\t]/gm, " ");
            topicName.textContent = name;
            const percent = nameToPercent.get(name);
            topicPercent.value = percent;
            percentMessage.textContent = percent.toFixed(2) + "% of all documents";
            const description = nameToDesc.get(name);
            console.log(description)
            if (description) {
                topicDesc.textContent = description;
            } else {
                topicDesc.textContent = "";
            }
            const keywords = nameToKeywords.get(name);
            console.log(keywords)
            if (keywords) {
                topicKeywords.textContent = "Keywords: " + keywords;
            } else {
                topicKeywords.textContent = "";
            }
        });
    });
}, 200);
    """
    plot = datamapplot.create_interactive_plot(
        coordinates,
        labels,
        font_family=font_family,
        logo="https://x-tabdeveloping.github.io/turftopic/images/logo.svg",
        logo_width=80,
        enable_topic_tree=enable_topic_tree,
        topic_tree_kwds=topic_tree_kwds,
        cluster_boundary_polygons=cluster_boundary_polygons,
        cluster_boundary_line_width=cluster_boundary_line_width,
        polygon_alpha=polygon_alpha,
        custom_css=CUSTOM_CSS,
        custom_html=custom_html,
        custom_js=custom_js,
        **kwargs,
    )

    def show_fig():
        with tempfile.TemporaryDirectory() as temp_dir:
            file_name = Path(temp_dir).joinpath("fig.html")
            plot.save(file_name)
            webbrowser.open("file://" + str(file_name.absolute()), new=2)
            time.sleep(2)

    plot.show = show_fig
    plot.write_html = plot.save
    return plot



import csv
import io

import numpy as np
from sklearn.preprocessing import label_binarize


def safe_binarize(y, classes) -> np.ndarray:
    """label_binarize, but its behaviour stays consistent when the labels are {0,1}"""
    if set(classes) == {0, 1}:
        binary_vec = np.squeeze(label_binarize(y, classes=classes))
        negative_vec = np.zeros(binary_vec.shape[0], dtype=binary_vec.dtype)
        negative_vec[binary_vec == 0] = 1
        # Concatenating
        onehot = np.stack((binary_vec, negative_vec))
        return onehot.T
    else:
        return label_binarize(y, classes=classes)


def export_table(
    table: list[list[str]],
    format="csv",
) -> str:
    if format == "csv":
        output = io.StringIO()
        writer = csv.writer(output, quoting=csv.QUOTE_NONNUMERIC)
        for row in table:
            writer.writerow(row)
        return output.getvalue()
    if format == "latex":
        columns, *rows = table
        n_columns = len(columns)
        latex_column_format = " ".join(["l"] * n_columns)
        latex_columns = " & ".join(columns) + "\\\\"
        latex_rows = [" & ".join(row) + "\\\\" for row in rows]
        latex_table = [
            "\\begin{center}",
            f"\\begin{{tabular}}{{ {latex_column_format} }}",
            latex_columns,
            "\\hline",
            *latex_rows,
            "\\end{tabular}",
            "\\end{center}",
        ]
        return "\n".join(latex_table)
    if format == "markdown":
        columns, *rows = table
        n_columns = len(columns)
        separator = ["-"] * n_columns
        markdown_rows = [
            "|" + "|".join([" " + value + " " for value in row]) + "|"
            for row in [columns, separator, *rows]
        ]
        return "\n".join(markdown_rows)
    raise ValueError(
        f"Format '{format}' not supported for tables, please use 'markdown', 'latex' or 'csv'"
    )


def sanitize_for_html(text: str) -> str:
    """Sanitizes strings so they can be put into JS or HTML strings"""
    # Escaping special characters
    text = (
        text.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
        .replace("'", "&#39;")
    )
    # Removing unnecessary whitespace
    text = " ".join(text.split())
    return text



from __future__ import annotations

from typing import Literal

import numpy as np
import scipy.sparse as spr
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.metrics.pairwise import cosine_similarity


def cluster_centroid_distance(
    cluster_centroids: np.ndarray,
    vocab_embeddings: np.ndarray,
) -> np.ndarray:
    """Computes feature importances based on distances between
    topic vectors (cluster centroids) and term embeddings

    Parameters
    ----------
    cluster_centroids: np.ndarray
        Coordinates of cluster centroids of shape (n_topics, embedding_size)
    vocab_embeddings: np.ndarray
        Term embeddings of shape (vocab_size, embedding_size)

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    n_components = cluster_centroids.shape[0]
    n_vocab = vocab_embeddings.shape[0]
    components = np.full((n_components, n_vocab), np.nan)
    valid_centroids = np.all(np.isfinite(cluster_centroids), axis=1)
    similarities = cosine_similarity(
        cluster_centroids[valid_centroids], vocab_embeddings
    )
    components[valid_centroids, :] = similarities
    return components


def linear_classifier(
    doc_topic_matrix: np.ndarray,
    embeddings: np.ndarray,
    vocab_embeddings: np.ndarray,
) -> np.ndarray:
    """Computes feature importances based on embedding directions
    obtained with a linear classifier.

    Parameters
    ----------
    doc_topic_matrix: np.ndarray
        Document-topic matrix.
    embeddings: np.ndarray
        Document embeddings.
    vocab_embeddings: np.ndarray
        Term embeddings of shape (vocab_size, embedding_size)

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    labels = np.argmax(doc_topic_matrix, axis=1)
    model = LinearDiscriminantAnalysis().fit(embeddings, labels)
    components = cosine_similarity(model.coef_, vocab_embeddings)
    if len(set(labels)) == 2:
        # Binary is a special case
        components = np.concatenate([-components, components], axis=0)
    return components


def fighting_words(
    doc_topic_matrix: np.ndarray,
    doc_term_matrix: spr.csr_matrix,
    prior: float | Literal["corpus"] = "corpus",
) -> np.ndarray:
    """Computes feature importance using the *Fighting Words* algorithm.

    Parameters
    ----------
    doc_topic_matrix: np.ndarray
        Document-topic matrix of shape (n_documents, n_topics)
    doc_term_matrix: np.ndarray
        Document-term matrix of shape (n_documents, vocab_size)
    prior: float or "corpus", default "corpus"
        Dirichlet prior to use. When a float, it indicates the alpha
        parameter of a symmetric Dirichlet, if "corpus",
        word frequencies from the background corpus are used.

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    labels = np.argmax(doc_topic_matrix, axis=1)
    n_topics = doc_topic_matrix.shape[1]
    n_vocab = doc_term_matrix.shape[1]
    components = []
    if prior == "corpus":
        priors = np.ravel(np.asarray(doc_term_matrix.sum(axis=0)))
    else:
        priors = np.full(n_vocab, prior)
    a0 = np.sum(priors)  # prior * n_vocab
    for i_topic in range(n_topics):
        topic_freq = np.ravel(
            np.asarray(doc_term_matrix[labels == i_topic].sum(axis=0))
        )
        rest_freq = np.ravel(
            np.asarray(doc_term_matrix[labels != i_topic].sum(axis=0))
        )
        n1 = np.sum(topic_freq)
        n2 = np.sum(rest_freq)
        topic_logodds = np.log(
            (topic_freq + priors) / (n1 + a0 - topic_freq - priors)
        )
        rest_logodds = np.log(
            (rest_freq + priors) / (n2 + a0 - rest_freq - priors)
        )
        delta = topic_logodds - rest_logodds
        delta_var = 1 / (topic_freq + priors) + 1 / (rest_freq + priors)
        zscore = delta / np.sqrt(delta_var)
        components.append(zscore)
    return np.stack(components)


def soft_ctf_idf(
    doc_topic_matrix: np.ndarray,
    doc_term_matrix: spr.csr_matrix,
    return_idf: bool = False,
) -> np.ndarray:
    """Computes feature importances using Soft C-TF-IDF

    Parameters
    ----------
    doc_topic_matrix: np.ndarray
        Document-topic matrix of shape (n_documents, n_topics)
    doc_term_matrix: np.ndarray
        Document-term matrix of shape (n_documents, vocab_size)

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    eps = np.finfo(float).eps
    term_importance = doc_topic_matrix.T @ doc_term_matrix
    overall_in_topic = np.abs(term_importance).sum(axis=1)
    n_docs = len(doc_topic_matrix)
    tf = (term_importance.T / (overall_in_topic + eps)).T
    idf = np.log(n_docs / (np.abs(term_importance).sum(axis=0) + eps))
    ctf_idf = tf * idf
    idf_diag = spr.diags(
        idf,
        offsets=0,
        shape=(doc_term_matrix.shape[1], doc_term_matrix.shape[1]),
        format="csr",
        dtype=tf.dtype,
    )
    if not return_idf:
        return ctf_idf
    else:
        return ctf_idf, idf_diag


def ctf_idf(
    doc_topic_matrix: np.ndarray,
    doc_term_matrix: spr.csr_matrix,
    return_idf: bool = False,
) -> np.ndarray:
    """Computes feature importances using standard C-TF-IDF

    Parameters
    ----------
    doc_topic_matrix: np.ndarray
        Document-topic matrix of shape (n_documents, n_topics)
    doc_term_matrix: np.ndarray
        Document-term matrix of shape (n_documents, vocab_size)

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    labels = np.argmax(doc_topic_matrix, axis=1)
    n_topics = doc_topic_matrix.shape[1]
    components = []
    overall_freq = np.ravel(np.asarray(doc_term_matrix.sum(axis=0)))
    average = overall_freq.sum() / n_topics
    overall_freq[overall_freq == 0] = np.finfo(float).eps
    for i_topic in range(n_topics):
        freq = np.ravel(
            np.asarray(doc_term_matrix[labels == i_topic].sum(axis=0))
        )
        component = freq * np.log(1 + average / overall_freq)
        components.append(component)
    idf = np.log((average / overall_freq) + 1)
    idf_diag = spr.diags(
        idf,
        offsets=0,
        shape=(doc_term_matrix.shape[1], doc_term_matrix.shape[1]),
        format="csr",
        dtype=doc_term_matrix.dtype,
    )
    if not return_idf:
        return np.stack(components)
    else:
        return np.stack(components), idf_diag


def npmi(
    doc_topic_matrix: np.ndarray,
    doc_term_matrix: spr.csr_matrix,
    smoothing: int = 5,
) -> np.ndarray:
    """Uses normalized pointwise mutual information between
    clusters and terms to calculate term importance scores.

    To not underestimate individual words' occurrances (overfit),
    a smoothing term is added, which is mathematically equivalent
    to using the MAP estimate of a symmetric dirichlet-multinomial model.

    Parameters
    ----------
    doc_topic_matrix: np.ndarray
        Document-topic matrix of shape (n_documents, n_topics)
    doc_term_matrix: np.ndarray
        Document-term matrix of shape (n_documents, vocab_size)
    smoothing: int, default 5
        Alpha parameter of the symmetric Dirichlet-multinomial.
        Corresponds to assuming that each term and cluster
        occurred this many more times than the observed.

    Returns
    -------
    ndarray of shape (n_topics, vocab_size)
        Term importance matrix.
    """
    eps = np.finfo(float).eps
    p_w = np.squeeze(np.asarray(doc_term_matrix.sum(axis=0))) + smoothing
    p_w = p_w / p_w.sum()
    p_w[p_w <= 0] = eps
    p_t = doc_topic_matrix.sum(axis=0) + smoothing
    p_t = p_t / p_t.sum()
    labels = np.argmax(doc_topic_matrix, axis=1)
    p_wt = []
    for i in np.arange(doc_topic_matrix.shape[1]):
        _p_w = (
            np.squeeze(np.asarray(doc_term_matrix[labels == i].sum(axis=0)))
            + smoothing
        )
        _p_w = _p_w / _p_w.sum()
        _p_w[_p_w <= 0] = eps
        p_wt.append(_p_w)
    p_wt = np.stack(p_wt)
    log_p_wt = np.log2(p_wt)
    numerator = log_p_wt - np.log2(p_w)
    denominator = -(log_p_wt.T - np.log2(p_t)).T
    res = numerator / denominator
    return res



from typing import Optional, Union

import numpy as np
from scipy.ndimage import (
    binary_erosion,
    generate_binary_structure,
    maximum_filter,
)
from scipy.stats import gaussian_kde
from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.manifold import TSNE
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.mixture import GaussianMixture
from sklearn.mixture._gaussian_mixture import (
    _compute_precision_cholesky,
    _estimate_gaussian_parameters,
)

from turftopic.base import Encoder
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.models.gmm import GMM, LexicalWordImportance


def detect_peaks(image):
    # define an 8-connected neighborhood
    neighborhood = generate_binary_structure(2, 25)
    local_max = maximum_filter(image, footprint=neighborhood) == image
    # local_max is a mask that contains the peaks we are
    # looking for, but also the background.
    # In order to isolate the peaks we must remove the background from the mask.
    # we create the mask of the background
    background = image == 0
    # a little technicality: we must erode the background in order to
    # successfully subtract it form local_max, otherwise a line will
    # appear along the background border (artifact of the local maximum filter)
    eroded_background = binary_erosion(
        background, structure=neighborhood, border_value=1
    )
    # we obtain the final mask, containing only peaks,
    # by removing the background from the local_max mask (xor operation)
    detected_peaks = local_max ^ eroded_background
    return detected_peaks


class FixedMeanGaussianMixture(GaussianMixture):
    def _m_step(self, X, log_resp):
        # Skipping mean update
        self.weights_, _, self.covariances_ = _estimate_gaussian_parameters(
            X, np.exp(log_resp), self.reg_covar, self.covariance_type
        )
        self.weights_ /= self.weights_.sum()
        self.precisions_cholesky_ = _compute_precision_cholesky(
            self.covariances_, self.covariance_type
        )


class Peax(ClusterMixin, BaseEstimator):
    """Clustering model based on density peaks.

    Parameters
    ----------
    random_state: int, default None
        Random seed to use for fitting gaussian mixture to peaks.
    """

    def __init__(self, random_state: Optional[int] = None):
        self.random_state = random_state

    def fit(self, X, y=None):
        self.X_range = np.min(X), np.max(X)
        self.density = gaussian_kde(X.T, "scott")
        coord = np.linspace(*self.X_range, num=100)
        z = []
        for yval in coord:
            points = np.stack([coord, np.full(coord.shape, yval)]).T
            prob = np.exp(self.density.logpdf(points.T))
            z.append(prob)
        z = np.stack(z)
        peaks = detect_peaks(z.T)
        peak_ind = np.nonzero(peaks)
        peak_pos = np.stack([coord[peak_ind[0]], coord[peak_ind[1]]]).T
        weights = self.density.pdf(peak_pos.T)
        weights = weights / weights.sum()
        self.gmm_ = FixedMeanGaussianMixture(
            peak_pos.shape[0],
            means_init=peak_pos,
            weights_init=weights,
            random_state=self.random_state,
        )
        self.labels_ = self.gmm_.fit_predict(X)
        # Checking whether there are close to zero components
        is_zero = np.isclose(self.gmm_.weights_, 0)
        n_zero = np.sum(is_zero)
        if n_zero > 0:
            print(
                f"{n_zero} components have zero weight, removing them and refitting."
            )
        peak_pos = peak_pos[~is_zero]
        weights = self.gmm_.weights_[~is_zero]
        weights = weights / weights.sum()
        self.gmm_ = FixedMeanGaussianMixture(
            peak_pos.shape[0],
            means_init=peak_pos,
            weights_init=weights,
            random_state=self.random_state,
        )
        self.labels_ = self.gmm_.fit_predict(X)
        self.classes_ = np.sort(np.unique(self.labels_))
        self.means_ = self.gmm_.means_
        self.weights_ = self.gmm_.weights_
        return self.labels_

    @property
    def n_components(self) -> int:
        return self.gmm_.n_components

    def predict_proba(self, X):
        return self.gmm_.predict_proba(X)

    def score_samples(self, X):
        return self.density.logpdf(X.T)

    def score(self, X):
        return np.mean(self.score_samples(X))


class Topeax(GMM):
    """Topic model based on the Peax clustering algorithm.
    The algorithm discovers the number of topics automatically, and is based on GMM.

    Parameters
    ----------
    encoder: str or SentenceTransformer
        Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
    vectorizer: CountVectorizer, default None
        Vectorizer used for term extraction.
        Can be used to prune or filter the vocabulary.
    perplexity: int, default 50
        Number of neighbours to take into account when running TSNE.
    random_state: int, default None
        Random state to use so that results are exactly reproducible.

    """

    def __init__(
        self,
        encoder: Union[
            Encoder, str, MultimodalEncoder
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        perplexity: int = 50,
        random_state: Optional[int] = None,
    ):
        dimensionality_reduction = TSNE(
            2,
            metric="cosine",
            perplexity=perplexity,
            random_state=random_state,
        )
        super().__init__(
            n_components=0,
            encoder=encoder,
            vectorizer=vectorizer,
            dimensionality_reduction=dimensionality_reduction,
            random_state=random_state,
        )

    def estimate_components(
        self,
        feature_importance: Optional[LexicalWordImportance] = None,
        doc_topic_matrix=None,
        doc_term_matrix=None,
    ) -> np.ndarray:
        doc_topic_matrix = (
            doc_topic_matrix
            if doc_topic_matrix is not None
            else self.doc_topic_matrix
        )
        doc_term_matrix = (
            doc_term_matrix
            if doc_term_matrix is not None
            else self.doc_term_matrix
        )
        lexical_components = super().estimate_components(
            "npmi", doc_topic_matrix, doc_term_matrix
        )
        vocab = self.get_vocab()
        if getattr(self, "vocab_embeddings", None) is None or (
            self.vocab_embeddings.shape[0] != vocab.shape[0]
        ):
            self.vocab_embeddings = self.encode_documents(vocab)
        topic_embeddings = []
        for weight in doc_topic_matrix.T:
            topic_embeddings.append(
                np.average(self.embeddings, axis=0, weights=weight)
            )
        self.topic_embeddings = np.stack(topic_embeddings)
        semantic_components = cosine_similarity(
            self.topic_embeddings, self.vocab_embeddings
        )
        # Transforming to positive values from 0 to 1
        # Then taking geometric average of the two values
        self.components_ = np.sqrt(
            ((1 + lexical_components) / 2) * ((1 + semantic_components) / 2)
        )
        return self.components_

    def _init_model(self, n_components: int):
        mixture = Peax()
        return mixture



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import numpy as np
from rich.console import Console
from rich.tree import Tree

from turftopic.base import ContextualModel

COLOR_PER_LEVEL = [
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_green",
    "bright_red",
    "bright_yellow",
    "cyan",
    "magenta",
    "blue",
    "white",
]


def _tree_plot(hierarchy: TopicNode):
    """Plots hierarchy with Plotly as a Tree"""
    try:
        import igraph as ig
        import plotly.express as px
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "You will need to install plotly and igraph to use hierarchical plotting functionality."
        ) from e

    def word_table(node):
        entries = []
        words = node.get_words(top_k=10)
        for word, imp in words:
            entries.append(f"<b>{word}</b>: <i>{imp:.2f}</i>")
        return " <br> ".join(entries)

    def traverse(h, nodes, edges, tables, parent=None):
        nodes.append(h)
        tables.append(word_table(h))
        if parent is not None:
            edges.append([parent._path_str(), h._path_str()])
        if h.children is not None:
            for child in h.children:
                traverse(child, nodes, edges, tables, parent=h)

    nodes = []
    edges = []
    tables = []
    for child in hierarchy.children:
        traverse(child, nodes, edges, tables)
    node_names = [node._simple_desc for node in nodes]
    node_ids = [node._path_str() for node in nodes]
    node_to_idx = {node_id: idx for idx, node_id in enumerate(node_ids)}
    edges_idx = [
        [node_to_idx[start], node_to_idx[end]] for start, end in edges
    ]
    graph = ig.Graph(len(nodes), edges=edges_idx, directed=True)
    layout = graph.layout("rt")
    layout.rotate(-90)
    x, y = np.array(layout.coords).T
    xmin, xmax = np.min(x), np.max(x)
    xpad = (xmax - xmin) * 0.35
    # Mapping root nodes to colors
    color_scheme = px.colors.qualitative.Dark24
    root_nodes = [str(node.path[0]) for node in nodes]
    color_map = {
        root_node: color_scheme[i % len(color_scheme)]
        for i, root_node in enumerate(np.unique(root_nodes))
    }
    fig = px.scatter(x=x, y=y, text=node_names, template="plotly_white")
    fig = fig.update_traces(
        customdata=np.array([[table] for table in tables]),
        hovertemplate="<b>%{text}</b> <br> <br> %{customdata[0]}",
        marker=dict(color=[color_map[root_node] for root_node in root_nodes]),
    )
    fig = fig.update_traces(marker=dict(size=20))
    fig = fig.update_layout(
        margin=dict(l=0, r=0, t=0, b=0),
        font=dict(family="Roboto Mono"),
    )
    fig = fig.update_yaxes(
        showgrid=False,
        visible=False,
        zeroline=False,
    )
    fig = fig.update_xaxes(
        showgrid=False,
        visible=False,
        zeroline=False,
        range=(xmin - xpad, xmax + xpad),
    )
    for start, end in edges_idx:
        fig.add_shape(
            type="line",
            xref="x",
            yref="y",
            x0=x[start],
            y0=y[start],
            x1=x[end],
            y1=y[end],
            opacity=0.2,
        )
    return fig


@dataclass
class TopicNode:
    """Node for a topic in a topic hierarchy.

    Parameters
    ----------
    model: ContextualModel
        Underlying topic model, which the hierarchy is based on.
    path: tuple[int], default ()
        Path that leads to this node from the root of the tree.
    word_importance: ndarray of shape (n_vocab), default None
        Importance of each word in the vocabulary for given topic.
    document_topic_vector: ndarray of shape (n_documents), default None
        Importance of the topic in all documents in the corpus.
    children: list[TopicNode], default None
        List of subtopics within this topic.
    """

    model: ContextualModel
    path: tuple[int] = ()
    word_importance: Optional[np.ndarray] = None
    document_topic_vector: Optional[np.ndarray] = None
    children: Optional[list[TopicNode]] = None

    def _path_str(self):
        return ".".join([str(level_id) for level_id in self.path])

    @property
    def classes_(self):
        if self.children is None:
            raise AttributeError("TopicNode doesn't have children.")
        return np.array([child.path[-1] for child in self.children])

    @property
    def components_(self):
        if self.children is None:
            raise AttributeError("TopicNode doesn't have children.")
        return np.stack([child.word_importance for child in self.children])

    @classmethod
    def create_root(
        cls,
        model: ContextualModel,
        components: np.ndarray,
        document_topic_matrix: np.ndarray,
    ) -> TopicNode:
        """Creates root node from a topic models' components and topic importances in documents."""
        children = []
        n_components = components.shape[0]
        classes = getattr(model, "classes_", None)
        if classes is None:
            classes = np.arange(n_components)
        for topic_id, comp, doc_top in zip(
            classes, components, document_topic_matrix.T
        ):
            children.append(
                cls(
                    model,
                    path=(topic_id,),
                    word_importance=comp,
                    document_topic_vector=doc_top,
                    children=None,
                )
            )
        return cls(
            model,
            path=(),
            word_importance=None,
            document_topic_vector=None,
            children=children,
        )

    @property
    def level(self) -> int:
        """Indicates how deep down the hierarchy the topic is."""
        return len(self.path)

    def get_words(self, top_k: int = 10) -> list[tuple[str, float]]:
        """Returns top words and words importances for the topic.

        Parameters
        ----------
        top_k: int, default 10
            Number of top words to return.

        Returns
        -------
        list[tuple[str, float]]
            List of word, importance pairs.
        """
        if self.word_importance is None:
            return []
        vocab = self.model.get_vocab()
        most_important = np.argsort(-self.word_importance)[:top_k]
        words = vocab[most_important]
        imp = self.word_importance[most_important]
        return list(zip(words, imp))

    @property
    def description(self) -> str:
        """Returns a high level description of the topic with its path in the tree
        and top words."""
        if not len(self.path):
            path = "Root"
        else:
            path = str(
                self.path[-1]
            )  # ".".join([str(idx) for idx in self.path])
        words = []
        for word, imp in self.get_words(top_k=10):
            words.append(word)
        concat_words = ", ".join(words)
        color = COLOR_PER_LEVEL[min(self.level, len(COLOR_PER_LEVEL) - 1)]
        stylized = f"[{color} bold]{path}[/]: [italic]{concat_words}[/]"
        console = Console()
        with console.capture() as capture:
            console.print(stylized, end="")
        return capture.get()

    @property
    def _simple_desc(self) -> str:
        if not len(self.path):
            path = "Root"
        else:
            path = str(
                self.path[-1]
            )  # ".".join([str(idx) for idx in self.path])
        words = []
        for word, imp in self.get_words(top_k=5):
            words.append(word)
        concat_words = ", ".join(words)
        return f"{path}: {concat_words}"

    def _build_tree(
        self,
        tree: Tree = None,
        top_k: int = 10,
        max_depth: Optional[int] = None,
    ) -> Tree:
        if tree is None:
            tree = Tree(self.description)
        else:
            tree = tree.add(self.description)
        out_of_depth = (max_depth is not None) and (self.level >= max_depth)
        if out_of_depth:
            if self.children is not None:
                tree.add("...")
            return tree
        if self.children is not None:
            for child in self.children:
                child._build_tree(tree, max_depth=max_depth)
        return tree

    def print_tree(
        self,
        top_k: int = 10,
        max_depth: Optional[int] = None,
    ) -> None:
        """Print hierarchy in tree form.

        Parameters
        ----------
        top_k: int, default 10
            Number of words to print for each topic.
        max_depth: int, default None
            Maximum depth at which topics should be printed in the hierarchy.
            If None, the entire hierarchy is printed.
        """
        tree = self._build_tree(top_k=top_k, max_depth=max_depth)
        console = Console()
        console.print(tree)

    def __str__(self):
        tree = self._build_tree(top_k=10, max_depth=3)
        console = Console()
        with console.capture() as capture:
            console.print(tree)
        return capture.get()

    def __repr__(self):
        return str(self)

    def __getitem__(self, id_or_path: int):
        if self.children is None:
            raise IndexError(
                "Current node is a leaf and does not have children."
            )
        mapping = {
            topic_class: i_topic
            for i_topic, topic_class in enumerate(self.classes_)
        }
        return self.children[mapping[id_or_path]]

    def __iter__(self):
        return iter(self.children)

    def plot_tree(self):
        """Plots hierarchy as an interactive tree in Plotly."""
        return _tree_plot(self)

    def _append_path(self, path_prefix: int):
        self.path = (path_prefix, *self.path)
        if self.children is not None:
            for child in self.children:
                child._append_path(path_prefix)

    def copy(self, deep: bool = True) -> TopicNode:
        """Creates a copy of the given node.

        Parameters
        ----------
        deep: bool, default True
            Indicates whether the copy should be deep or shallow.
            Deep copies are done recursively, while shallow copies only
            contain references to the original children.

        Returns
        -------
        Copy of original hierarchy.
        """
        if (self.children is None) or (not deep):
            return type(self)(
                model=self.model,
                path=self.path,
                children=self.children,
                word_importance=self.word_importance,
                document_topic_vector=self.document_topic_vector,
            )
        else:
            children = [child.copy(deep=True) for child in self.children]
            return type(self)(
                model=self.model,
                path=self.path,
                children=children,
                word_importance=self.word_importance,
                document_topic_vector=self.document_topic_vector,
            )

    def cut(self, max_depth: int) -> TopicNode:
        """Cuts hierarchy at a given depth, returns copy of the hierarchy with levels beyond max_depth removed.

        Parameters
        ----------
        max_depth: int
            Maximum level of nodes to keep.

        Returns
        -------
        TopicNode
            Hierarchy cut at the given level.
            Contains a deep copy of the original nodes.
        """
        if (self.level >= max_depth) or (not self.children):
            return type(self)(
                model=self.model,
                path=self.path,
                children=None,
                word_importance=self.word_importance,
                document_topic_vector=self.document_topic_vector,
            )
        else:
            children = [child.cut(max_depth) for child in self.children]
            return type(self)(
                model=self.model,
                path=self.path,
                children=children,
                word_importance=self.word_importance,
                document_topic_vector=self.document_topic_vector,
            )

    def collect_leaves(self) -> list[TopicNode]:
        def _collect_leaves(node: TopicNode, leaves: list[TopicNode]):
            if not node.children:
                leaves.append(node.copy(deep=False))
            else:
                for child in node.children:
                    _collect_leaves(child, leaves)

        leaves = []
        _collect_leaves(self, leaves)
        return leaves

    def flatten(self) -> TopicNode:
        """Returns new hierarchy with only the leaves of the tree.

        Returns
        -------
        TopicNode
            Root node containing all leaves in a hierarchy.
            Copies of the original nodes.
        """
        leaves = self.collect_leaves()
        ids = [leaf.path[-1] for leaf in leaves]
        # If the IDs are not unique, we label them from 0 to N
        if len(set(ids)) != len(ids):
            current = 0
            new_ids = []
            for node_id in ids:
                if node_id != -1:
                    new_ids.append(current)
                    current += 1
                else:
                    new_ids.append(-1)
            ids = new_ids
        for leaf_id, leaf in zip(ids, leaves):
            leaf.path = (*self.path, leaf_id)
        return type(self)(
            model=self.model,
            path=self.path,
            word_importance=self.word_importance,
            document_topic_vector=self.document_topic_vector,
            children=leaves,
        )


@dataclass
class DivisibleTopicNode(TopicNode):
    """Node for a topic in a topic hierarchy that can be subdivided."""

    def clear(self):
        """Deletes children of the given node."""
        self.children = None
        return self

    def divide(self, n_subtopics: int, **kwargs):
        """Divides current node into smaller subtopics.
        Only works when the underlying model is a divisive hierarchical model.

        Parameters
        ----------
        n_subtopics: int
            Number of topics to divide the topic into.
        """
        try:
            self.children = self.model.divide_topic(
                node=self, n_subtopics=n_subtopics, **kwargs
            )
        except AttributeError as e:
            raise AttributeError(
                "Looks like your model is not a divisive hierarchical model."
            ) from e
        return self

    def divide_children(self, n_subtopics: int, **kwargs):
        """Divides all children of the current node to smaller topics.
        Only works when the underlying model is a divisive hierarchical model.

        Parameters
        ----------
        n_subtopics: int
            Number of topics to divide the topics into.
        """
        if self.children is None:
            raise ValueError(
                "Current Node is a leaf, children can't be subdivided."
            )
        for child in self.children:
            child.divide(n_subtopics, **kwargs)
        return self

    def __str__(self):
        tree = self._build_tree(top_k=10, max_depth=3)
        console = Console()
        with console.capture() as capture:
            console.print(tree)
        return capture.get()

    def __repr__(self):
        return str(self)



from typing import Optional

import openai

from turftopic.analyzers.base import Analyzer


class OpenAIAnalyzer(Analyzer):
    """Analyze topic model with an OpenAI LLM.

    Parameters
    ----------
    model_name: str, default 'gpt-5-nano'
        OpenAI model to use.
    use_summaries: bool, default False
        Indicates whether the language model should summarize documents before
        analyzing the topics.
    context: str, default None
        Additional context provided to the analyzer for analysis.
        e.g. "Analyze topics from blog posts related to morality and religion"
    system_prompt: str, default None
        System prompt to use for the language model.
    summary_prompt: str, default None
        Prompt to use for abstractive summarization.
    namer_prompt: str, default None
        Prompt template for naming topics.
    description_prompt: str, default None
        Prompt template for generating topic descriptions.
    """

    def __init__(
        self,
        model_name: str = "gpt-5-nano",
        context: Optional[str] = None,
        use_summaries: bool = False,
        system_prompt: Optional[str] = None,
        summary_prompt: Optional[str] = None,
        namer_prompt: Optional[str] = None,
        description_prompt: Optional[str] = None,
    ):
        self.client = openai.OpenAI()
        self.model_name = model_name
        self.system_prompt = system_prompt or self.system_prompt
        self.summary_prompt = summary_prompt or self.summary_prompt
        self.namer_prompt = namer_prompt or self.namer_prompt
        self.description_prompt = description_prompt or self.description_prompt
        self.use_summaries = use_summaries

    def generate_text(self, prompt: str) -> str:
        messages = [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": prompt},
        ]
        response = self.client.chat.completions.create(
            messages=messages,
            model=self.model_name,
        )
        return response.choices[0].message.content



import typing
from typing import Literal

DistanceMetric = Literal[
    "cityblock",
    "cosine",
    "euclidean",
    "l1",
    "l2",
    "manhattan",
    "braycurtis",
    "canberra",
    "chebyshev",
    "correlation",
    "dice",
    "hamming",
    "jaccard",
    "kulsinski",
    "mahalanobis",
    "minkowski",
    "rogerstanimoto",
    "russellrao",
    "seuclidean",
    "sokalmichener",
    "sokalsneath",
    "sqeuclidean",
    "yule",
]
VALID_DISTANCE_METRICS = list(typing.get_args(DistanceMetric))



import warnings
from datetime import datetime
from functools import partial
from typing import Literal, Optional, Union

import numpy as np
from rich.console import Console
from sentence_transformers import SentenceTransformer
from sklearn.base import TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.mixture import BayesianGaussianMixture, GaussianMixture
from sklearn.pipeline import Pipeline

from turftopic._datamapplot import build_datamapplot
from turftopic.base import ContextualModel, Encoder
from turftopic.dynamic import DynamicTopicModel
from turftopic.encoders.multimodal import MultimodalEncoder
from turftopic.feature_importance import (
    ctf_idf,
    fighting_words,
    npmi,
    soft_ctf_idf,
)
from turftopic.multimodal import (
    Image,
    ImageRepr,
    MultimodalEmbeddings,
    MultimodalModel,
)
from turftopic.optimization import optimize_n_components
from turftopic.vectorizers.default import default_vectorizer

FEATURE_IMPORTANCE_METHODS = {
    "soft-c-tf-idf": soft_ctf_idf,
    "c-tf-idf": ctf_idf,
    "fighting-words": fighting_words,
    "npmi": partial(npmi, smoothing=2),
}
LexicalWordImportance = Literal[
    "soft-c-tf-idf",
    "c-tf-idf",
    "npmi",
    "fighting-words",
]


class GMM(ContextualModel, DynamicTopicModel, MultimodalModel):
    """Multivariate Gaussian Mixture Model over document embeddings.
        Models topics as mixture components.

        ```python
        from turftopic import GMM
    corpus: list[str] = ["some text", "more text", ...]

        model = GMM(10, weight_prior="dirichlet_process").fit(corpus)
        model.print_topics()
        ```

        Parameters
        ----------
        n_components: int or "auto"
            Number of topics.
            If "auto", the Bayesian Information criterion
            will be used to estimate this quantity.
            *Note that "auto" can only be used when no priors as specified*.
        encoder: str or SentenceTransformer
            Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
        vectorizer: CountVectorizer, default None
            Vectorizer used for term extraction.
            Can be used to prune or filter the vocabulary.
        weight_prior: 'dirichlet', 'dirichlet_process' or None, default 'dirichlet'
            Prior to impose on component weights, if None,
            maximum likelihood is optimized with expectation maximization,
            otherwise variational inference is used.
        gamma: float, default None
            Concentration parameter of the symmetric prior.
            By default 1/n_components is used.
            Ignored when weight_prior is None.
        dimensionality_reduction: TransformerMixin, default None
            Optional dimensionality reduction step before GMM is run.
            This is recommended for very large datasets with high dimensionality,
            as the number of parameters grows vast in the model otherwise.
            We recommend using PCA, as it is a linear solution, and will likely
            result in Gaussian components.
            For even larger datasets you can use IncrementalPCA to reduce
            memory load.
        feature_importance: LexicalWordImportance, default 'soft-c-tf-idf'
            Feature importance method to use.
            *Note that only lexical methods can be used with GMM,
            not embedding-based ones.*
        random_state: int, default None
            Random state to use so that results are exactly reproducible.

        Attributes
        ----------
        weights_: ndarray of shape (n_components)
            Weights of the different mixture components.
    """

    def __init__(
        self,
        n_components: Union[int, Literal["auto"]],
        encoder: Union[
            Encoder, str, MultimodalEncoder
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        dimensionality_reduction: Optional[TransformerMixin] = None,
        feature_importance: LexicalWordImportance = "soft-c-tf-idf",
        weight_prior: Literal["dirichlet", "dirichlet_process", None] = None,
        gamma: Optional[float] = None,
        random_state: Optional[int] = None,
    ):
        self.n_components = n_components
        self.encoder = encoder
        self.weight_prior = weight_prior
        self.gamma = gamma
        self.random_state = random_state
        if isinstance(encoder, str):
            self.encoder_ = SentenceTransformer(encoder)
        else:
            self.encoder_ = encoder
        self.validate_encoder()
        if vectorizer is None:
            self.vectorizer = default_vectorizer()
        else:
            self.vectorizer = vectorizer
        if feature_importance not in FEATURE_IMPORTANCE_METHODS:
            valid = list(FEATURE_IMPORTANCE_METHODS.keys())
            raise ValueError(
                f"{feature_importance} not in list of valid feature importance methods: {valid}"
            )
        self.feature_importance = feature_importance
        self.dimensionality_reduction = dimensionality_reduction
        if (self.n_components == "auto") and (self.weight_prior is not None):
            raise ValueError(
                "You cannot use N='auto' with a prior. Try setting weight_prior=None."
            )

    def estimate_components(
        self,
        feature_importance: Optional[LexicalWordImportance] = None,
        doc_topic_matrix=None,
        doc_term_matrix=None,
    ) -> np.ndarray:
        feature_importance = feature_importance or self.feature_importance
        imp_fn = FEATURE_IMPORTANCE_METHODS[feature_importance]
        doc_topic_matrix = (
            doc_topic_matrix
            if doc_topic_matrix is not None
            else self.doc_topic_matrix
        )
        doc_term_matrix = (
            doc_term_matrix
            if doc_term_matrix is not None
            else self.doc_term_matrix
        )
        self.components_ = imp_fn(doc_topic_matrix, doc_term_matrix)
        return self.components_

    def _create_bic(self, embeddings: np.ndarray):
        def f_bic(n_components: int):
            random_state = 42
            success = False
            n_tries = 1
            while not success and (n_tries <= 5):
                try:
                    # This can sometimes run into problems especially
                    # with covariance estimation
                    model = GaussianMixture(
                        n_components, random_state=self.random_state
                    )
                    model.fit(embeddings)
                    success = True
                except Exception:
                    random_state += 1
                    n_tries += 1
            if n_tries > 5:
                return 0
            return model.bic(embeddings)

        return f_bic

    def _init_model(self, n_components: int):
        if self.weight_prior is not None:
            mixture = BayesianGaussianMixture(
                n_components=n_components,
                weight_concentration_prior_type=(
                    "dirichlet_distribution"
                    if self.weight_prior == "dirichlet"
                    else "dirichlet_process"
                ),
                weight_concentration_prior=self.gamma,
                random_state=self.random_state,
            )
        else:
            mixture = GaussianMixture(
                n_components, random_state=self.random_state
            )
        return mixture

    def fit_transform(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        console = Console()
        with console.status("Fitting model") as status:
            if embeddings is None:
                status.update("Encoding documents")
                embeddings = self.encoder_.encode(raw_documents)
                console.log("Documents encoded.")
            self.embeddings = embeddings
            status.update("Extracting terms.")
            self.doc_term_matrix = self.vectorizer.fit_transform(raw_documents)
            console.log("Term extraction done.")
            X = embeddings
            if self.dimensionality_reduction is not None:
                status.update("Reducing embedding dimensionality.")
                X = self.dimensionality_reduction.fit_transform(embeddings)
                console.log("Dimensionality reduction complete.")
                self.reduced_embeddings = X
            n_components = self.n_components
            if self.n_components == "auto":
                status.update("Finding optimal value of N")
                f_bic = self._create_bic(X)
                n_components = optimize_n_components(f_bic, verbose=True)
                console.log(f"Found optimal N={n_components}.")
            status.update("Fitting mixture model.")
            self.gmm_ = self._init_model(n_components)
            self.gmm_.fit(X)
            console.log("Mixture model fitted.")
            status.update("Estimating term importances.")
            self.doc_topic_matrix = self.gmm_.predict_proba(X)
            self.components_ = self.estimate_components()
            console.log("Model fitting done.")
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=self.doc_topic_matrix
            )
        return self.doc_topic_matrix

    def fit_transform_multimodal(
        self,
        raw_documents: list[str],
        images: list[ImageRepr],
        y=None,
        embeddings: Optional[MultimodalEmbeddings] = None,
    ) -> np.ndarray:
        self.validate_embeddings(embeddings)
        console = Console()
        self.multimodal_embeddings = embeddings
        with console.status("Fitting model") as status:
            if self.multimodal_embeddings is None:
                status.update("Encoding documents")
                self.multimodal_embeddings = self.encode_multimodal(
                    raw_documents, images
                )
                console.log("Documents encoded.")
            status.update("Extracting terms.")
            self.doc_term_matrix = self.vectorizer.fit_transform(raw_documents)
            console.log("Term extraction done.")
            X = self.multimodal_embeddings["document_embeddings"]
            if self.dimensionality_reduction is not None:
                status.update("Reducing embedding dimensionality.")
                X = self.dimensionality_reduction.fit_transform(embeddings)
                console.log("Dimensionality reduction complete.")
            n_components = self.n_components
            if self.n_components == "auto":
                status.update("Finding optimal value of N")
                f_bic = self._create_bic(X)
                n_components = optimize_n_components(f_bic, verbose=True)
                console.log(f"Found optimal N={n_components}.")
            status.update("Fitting mixture model.")
            self.gmm_ = self._init_model(n_components)
            self.gmm_.fit(X)
            console.log("Mixture model fitted.")
            status.update("Estimating term importances.")
            self.doc_topic_matrix = self.gmm_.predict_proba(X)
            self.components_ = self.estimate_components()
            console.log("Model fitting done.")
            try:
                self.image_topic_matrix = self.transform(
                    raw_documents,
                    embeddings=self.multimodal_embeddings["image_embeddings"],
                )
            except Exception as e:
                warnings.warn(
                    f"Couldn't produce image topic matrix due to exception: {e}, using doc-topic matrix."
                )
                self.image_topic_matrix = self.doc_topic_matrix
            self.top_images: list[list[Image.Image]] = self.collect_top_images(
                images, self.image_topic_matrix
            )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=self.doc_topic_matrix
            )
            console.log("Transformation done.")
        return self.doc_topic_matrix

    @property
    def labels_(self):
        return np.argmax(self.doc_topic_matrix, axis=1)

    @property
    def weights_(self) -> np.ndarray:
        if isinstance(self.gmm_, Pipeline):
            model = self.gmm_.steps[-1][1]
        else:
            model = self.gmm_
        return model.weights_

    def transform(
        self, raw_documents, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Infers topic importances for new documents based on a fitted model.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.

        Returns
        -------
        ndarray of shape (n_dimensions, n_topics)
            Document-topic matrix.
        """
        if embeddings is None:
            embeddings = self.encoder_.encode(raw_documents)
        if self.dimensionality_reduction is not None:
            embeddings = self.dimensionality_reduction.transform(embeddings)
        return self.gmm_.predict_proba(embeddings)

    def fit_transform_dynamic(
        self,
        raw_documents,
        timestamps: list[datetime],
        embeddings: Optional[np.ndarray] = None,
        bins: Union[int, list[datetime]] = 10,
    ):
        time_labels, self.time_bin_edges = self.bin_timestamps(
            timestamps, bins
        )
        if hasattr(self, "components_"):
            doc_topic_matrix = self.transform(
                raw_documents, embeddings=embeddings
            )
        else:
            doc_topic_matrix = self.fit_transform(
                raw_documents, embeddings=embeddings
            )
        self.doc_term_matrix = self.vectorizer.transform(raw_documents)
        n_comp, n_vocab = self.components_.shape
        n_bins = len(self.time_bin_edges) - 1
        self.temporal_components_ = np.zeros(
            (n_bins, n_comp, n_vocab), dtype=self.doc_term_matrix.dtype
        )
        self.temporal_importance_ = np.zeros((n_bins, n_comp))
        for i_timebin in np.unique(time_labels):
            topic_importances = doc_topic_matrix[time_labels == i_timebin].sum(
                axis=0
            )
            # Normalizing
            topic_importances = topic_importances / topic_importances.sum()
            components = self.estimate_components(
                doc_topic_matrix=doc_topic_matrix[time_labels == i_timebin],
                doc_term_matrix=self.doc_term_matrix[time_labels == i_timebin],  # type: ignore
            )
            self.temporal_components_[i_timebin] = components
            self.temporal_importance_[i_timebin] = topic_importances
        return doc_topic_matrix

    def plot_components_datamapplot(
        self,
        coordinates: Optional[np.ndarray] = None,
        hover_text: Optional[list[str]] = None,
        **kwargs,
    ):
        """Creates an interactive browser plot of the topics in your data using datamapplot.

        Parameters
        ----------
        coordinates: np.ndarray, default None
            Lower dimensional projection of the embeddings.
            If None, will try to use the projections from the
            dimensionality_reduction method of the model.
        hover_text: list of str, optional
            Text to show when hovering over a document.

        Returns
        -------
        plot
            Interactive datamap plot, you can call the `.show()` method to
            display it in your default browser or save it as static HTML using `.write_html()`.
        """
        if coordinates is None:
            if not hasattr(self, "reduced_embeddings"):
                raise ValueError(
                    "Coordinates not specified, but the model does not contain reduced embeddings."
                )
            coordinates = self.reduced_embeddings[:, (0, 1)]
        labels = np.argmax(self.doc_topic_matrix, axis=1)
        plot = build_datamapplot(
            coordinates=coordinates,
            topic_names=self.topic_names,
            labels=labels,
            classes=np.arange(self.gmm_.n_components),
            top_words=self.get_top_words(),
            hover_text=hover_text,
            topic_descriptions=getattr(self, "topic_descriptions", None),
            **kwargs,
        )
        return plot

    def plot_density(
        self, hover_text: list[str] = None, show_points=False, light_mode=False
    ):
        try:
            import plotly.graph_objects as go
        except (ImportError, ModuleNotFoundError) as e:
            raise ModuleNotFoundError(
                "Please install plotly if you intend to use plots in Turftopic."
            ) from e

        if not hasattr(self, "reduced_embeddings"):
            raise ValueError(
                "No reduced embeddings found, can't display in 2d space."
            )
        if self.reduced_embeddings.shape[1] != 2:
            warnings.warn(
                "Embeddings are not in 2d space, only using first 2 dimensions"
            )

        coord_min, coord_max = np.min(self.reduced_embeddings), np.max(
            self.reduced_embeddings
        )
        coord_spread = coord_max - coord_min
        coord_min = coord_min - coord_spread * 0.05
        coord_max = coord_max + coord_spread * 0.05
        coord = np.linspace(coord_min, coord_max, num=100)
        z = []
        for yval in coord:
            points = np.stack([coord, np.full(coord.shape, yval)]).T
            prob = np.exp(self.gmm_.score_samples(points))
            z.append(prob)
        z = np.stack(z)
        color_grid = [0.0, 0.25, 0.5, 0.75, 1.0]
        colorscale = [
            "#01014B",
            "#000080",
            "#5D5DEF",
            "#B7B7FF",
            "#ffffff",
        ]
        if light_mode:
            colorscale = colorscale[::-1]
        traces = [
            go.Contour(
                z=z,
                colorscale=list(zip(color_grid, colorscale)),
                showscale=False,
                x=coord,
                y=coord,
                hoverinfo="skip",
            ),
        ]
        if show_points:
            scatter = go.Scatter(
                x=self.reduced_embeddings[:, 0],
                y=self.reduced_embeddings[:, 1],
                mode="markers",
                showlegend=False,
                text=hover_text,
                marker=dict(
                    symbol="circle",
                    opacity=0.5,
                    color="white",
                    size=8,
                    line=dict(width=1),
                ),
            )
            traces.append(scatter)
        fig = go.Figure(data=traces)
        fig = fig.update_layout(
            showlegend=False, margin=dict(r=0, l=0, t=0, b=0)
        )
        fig = fig.update_xaxes(showticklabels=False)
        fig = fig.update_yaxes(showticklabels=False)
        for mean, name, keywords in zip(
            self.gmm_.means_, self.topic_names, self.get_top_words()
        ):
            _keys = ""
            for i, key in enumerate(keywords):
                if (i % 5) == 0:
                    _keys += "<br> "
                _keys += key
                if i < (len(keywords) - 1):
                    _keys += ","
                _keys += " "
            text = f"<b>{name}</b> <i>{_keys}</i> "
            fig.add_annotation(
                text=text,
                x=mean[0],
                y=mean[1],
                align="left",
                showarrow=False,
                xshift=0,
                yshift=50,
                font=dict(family="Roboto Mono", size=18, color="black"),
                bgcolor="rgba(255,255,255,0.9)",
                bordercolor="black",
                borderwidth=2,
            )
        return fig



import numpy as np
from sklearn.feature_extraction.text import CountVectorizer

try:
    import snowballstemmer
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "Snowball is not installed on your machine, please run `pip install turftopic[snowball]` before using the snowball stemmer."
    ) from e


class StemmingCountVectorizer(CountVectorizer):
    """Extractes stemmed words from documents using Snowball."""

    def __init__(
        self,
        language="english",
        *,
        input="content",
        encoding="utf-8",
        decode_error="strict",
        strip_accents=None,
        lowercase=True,
        preprocessor=None,
        tokenizer=None,
        stop_words=None,
        token_pattern=r"(?u)\b\w\w+\b",
        ngram_range=(1, 1),
        analyzer="word",
        max_df=1.0,
        min_df=1,
        max_features=None,
        vocabulary=None,
        binary=False,
        dtype=np.int64,
    ):
        self.language = language
        self._stemmer = snowballstemmer.stemmer(self.language)
        super().__init__(
            input=input,
            encoding=encoding,
            decode_error=decode_error,
            strip_accents=strip_accents,
            lowercase=lowercase,
            preprocessor=preprocessor,
            tokenizer=tokenizer,
            stop_words=stop_words,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
            analyzer=analyzer,
            max_df=max_df,
            min_df=min_df,
            max_features=max_features,
            vocabulary=vocabulary,
            binary=binary,
            dtype=dtype,
        )

    def build_tokenizer(self):
        super_tokenizer = super().build_tokenizer()

        def tokenizer(text):
            return self._stemmer.stemWords(super_tokenizer(text))

        return tokenizer



from typing import Protocol

from PIL import Image


class MultimodalEncoder(Protocol):
    """Base class for external encoder models."""

    def get_text_embeddings(
        self,
        texts: list[str],
        *,
        batch_size: int = 8,
        **kwargs,
    ): ...

    def get_image_embeddings(
        self,
        images: list[Image.Image],
        *,
        batch_size: int = 8,
        **kwargs,
    ): ...

    def get_fused_embeddings(
        self,
        texts: list[str] = None,
        images: list[Image.Image] = None,
        batch_size: int = 8,
        **kwargs,
    ): ...



import numpy as np
from sklearn.decomposition._nmf import _beta_divergence, _initialize_nmf
from sklearn.utils.extmath import safe_sparse_dot

EPSILON = np.finfo(np.float32).eps


def weighted_nmf(
    dtm: np.ndarray,
    weight: np.ndarray,
    n_components: int,
    seed: int,
    max_iter: int = 200,
    tol: float = 1e-4,
) -> tuple[np.ndarray, np.ndarray]:
    """Multiplicative Update algorithm for a special case of weighted NMF, where
    only the rows are weighted, but not the individual elements in the data matrix."""
    doc_topic_matrix, components = _initialize_nmf(
        dtm, n_components, random_state=seed
    )
    U = components.T
    V = doc_topic_matrix.T
    weighted_A = dtm.T.multiply(weight)  # .T
    prev_error = np.inf
    for i in range(0, max_iter):
        # Update V
        numerator = safe_sparse_dot(U.T, weighted_A)
        denominator = np.linalg.multi_dot((U.T, U, V * weight))
        denominator[denominator <= 0] = EPSILON
        delta = numerator
        delta /= denominator
        delta[np.isinf(delta) & (V == 0)] = 0
        V *= delta
        # Update U
        numerator = safe_sparse_dot(weighted_A, V.T)
        denominator = np.linalg.multi_dot((U, V * weight, V.T))
        denominator[denominator <= 0] = EPSILON
        delta = numerator
        delta /= denominator
        delta[np.isinf(delta) & (U == 0)] = 0
        U *= delta
        if (tol > 0) and (i % 10 == 0):
            error = _beta_divergence(dtm, V.T, U.T, 2)
            if (error - prev_error) > tol:
                break
            prev_error = error
    components, doc_topic_matrix = U.T, V.T
    return components, doc_topic_matrix



import json
import warnings
from pathlib import Path
from typing import Union

import joblib
import pkg_resources
from huggingface_hub import snapshot_download

IMPORTANT_PACKAGES = [
    "scikit-learn",
    "sentence-transformers",
    "turftopic",
    "joblib",
]

README_TEMPLATE = """
---
pipeline_tag: text-classification
library_name: turftopic
tags:
- turftopic
- topic-modelling
---

# {model_path}

This repository contains a topic model trained with the [Turftopic](https://github.com/x-tabdeveloping/turftopic) Python library.

To load and use the model run the following piece of code:

```python
from turftopic import load_model

model = load_model({model_path})
model.print_topics()
```

## Model Structure

The model is structured as follows:

```
{model_structure}
```

## Topics
The topics discovered by the model are the following:

{topics_table}

## Package versions

The model in this repo was trained using the following package versions:

{package_version_table}

We recommend that you install the same, or compatible versions of these packages locally, before trying to load a model.

"""


def get_package_versions() -> dict[str, str]:
    return {
        package: pkg_resources.get_distribution(package).version
        for package in IMPORTANT_PACKAGES
    }


def validate_package_versions(remote_versions: dict[str, str]):
    local_versions = get_package_versions()
    for package in IMPORTANT_PACKAGES:
        local = local_versions.get(package, None)
        remote = remote_versions.get(package, None)
        if local != remote:
            warnings.warn(
                f"Local version of {package} ({local}) does not match the version in the model you are trying to load ({remote}). This might cause some issues."
            )


def create_readme(model, model_path: str) -> str:
    model_structure = str(model)
    topics_table = model.export_topics(format="markdown", top_k=10)
    local_versions = get_package_versions()
    lines = ["| Package | Version |", "| - | - |"]
    for package in IMPORTANT_PACKAGES:
        version = local_versions.get(package, "")
        lines.append(f"| {package} | {version} |")
    package_version_table = "\n".join(lines)
    return README_TEMPLATE.format(
        model_path=model_path,
        model_structure=model_structure,
        topics_table=topics_table,
        package_version_table=package_version_table,
    )


def load_model(repo_id_or_path: Union[str, Path]):
    """Loads topic model from local directory or HuggingFace Hub repository.

    Parameters
    ----------
    repo_id_or_path: str | Path
        Path to local directory or HuggingFace repository ID.
    """
    path = Path(repo_id_or_path)
    if path.is_dir():
        with path.joinpath("package_versions.json").open() as ver_file:
            remote_versions = json.loads(ver_file.read())
        validate_package_versions(remote_versions)
        model = joblib.load(path.joinpath("model.joblib"))
        return model
    else:
        in_dir = snapshot_download(repo_id=repo_id_or_path)
        return load_model(in_dir)



import math
from typing import Optional, Union

import numpy as np
import torch
from rich.console import Console
from sentence_transformers import SentenceTransformer
from sklearn.feature_extraction.text import CountVectorizer

from turftopic.base import ContextualModel, Encoder
from turftopic.models._fastopic import fastopic
from turftopic.vectorizers.default import default_vectorizer


class FASTopic(ContextualModel):
    """
    Implementation of the FASTopic model with a Turftopic API.
    The implementation is based on the [original FASTopic package](https://github.com/BobXWu/FASTopic/tree/master),
    but is adapted for optimal use in Turftopic (you can pre-compute embeddings for instance).

    You will need to install torch to use this model.

    ```bash
    pip install turftopic[torch]
    ## OR:
    pip install turftopic[pyro-ppl]
    ```

    ```python
    from turftopic import FASTopic

    corpus: list[str] = ["some text", "more text", ...]

    model = FASTopic(10).fit(corpus)
    model.print_topics()
    ```

    Parameters
    ----------
    n_components: int
        Number of topics. If you're using priors on the weight,
        feel free to overshoot with this value.
    encoder: str or SentenceTransformer
        Model to encode documents/terms, all-MiniLM-L6-v2 is the default.
    vectorizer: CountVectorizer, default None
        Vectorizer used for term extraction.
        Can be used to prune or filter the vocabulary.
    random_state: int, default None
        Random state to use so that results are exactly reproducible.
    DT_alpha: float, default 3.0
        Sinkhorn alpha between document embeddings and topic embeddings.
    TW_alpha: float, default 2.0
        Sinkhorn alpha between topic embeddings and word embeddings.
    theta_temp: float, default 1.0
        Temperature parameter of used in softmax to compute topic probabilities in documents.
    n_epochs: int, default 200
        Number of epochs to train the model for.
    learning_rate: float, default 0.002
        Learning rate for the ADAM optimizer.
    device: str, default "cpu"
        Device to run the model on. Defaults to CPU.
    """

    def __init__(
        self,
        n_components: int,
        encoder: Union[
            Encoder, str
        ] = "sentence-transformers/all-MiniLM-L6-v2",
        vectorizer: Optional[CountVectorizer] = None,
        random_state: Optional[int] = None,
        batch_size: Optional[int] = None,
        DT_alpha: float = 3.0,
        TW_alpha: float = 2.0,
        theta_temp: float = 1.0,
        n_epochs: int = 200,
        learning_rate: float = 0.002,
        device: str = "cpu",
    ):
        self.n_components = n_components
        self.encoder = encoder
        self.random_state = random_state
        if isinstance(encoder, str):
            self.encoder_ = SentenceTransformer(encoder)
        else:
            self.encoder_ = encoder
        if vectorizer is None:
            self.vectorizer = default_vectorizer()
        else:
            self.vectorizer = vectorizer
        self.DT_alpha = DT_alpha
        self.TW_alpha = TW_alpha
        self.theta_temp = theta_temp
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.device = device
        self.model = fastopic(n_components, theta_temp, DT_alpha, TW_alpha)
        self.batch_size = batch_size

    def make_optimizer(self, learning_rate: float):
        args_dict = {
            "params": self.model.parameters(),
            "lr": learning_rate,
        }
        optimizer = torch.optim.Adam(**args_dict)
        return optimizer

    def _train_model(self, embeddings, document_term_matrix, status):
        self.model.init(document_term_matrix.shape[1], embeddings.shape[1])
        self.model = self.model.to(self.device)
        optimizer = self.make_optimizer(self.learning_rate)
        self.model.train()
        if self.batch_size is None:
            batch_size = embeddings.shape[0]
        else:
            batch_size = self.batch_size
        num_batches = int(
            math.ceil(document_term_matrix.shape[0] / batch_size)
        )
        for epoch in range(self.n_epochs):
            running_loss = 0
            for i in range(num_batches):
                batch_bow = np.atleast_2d(
                    document_term_matrix[
                        i * batch_size : (i + 1) * batch_size, :
                    ].toarray()
                )
                # Skipping batches that are smaller than 2
                if batch_bow.shape[0] < 2:
                    continue
                batch_contextualized = np.atleast_2d(
                    embeddings[i * batch_size : (i + 1) * batch_size, :]
                )
                batch_contextualized = (
                    torch.tensor(batch_contextualized).float().to(self.device)
                )
                batch_bow = torch.tensor(batch_bow).float().to(self.device)
                rst_dict = self.model(batch_bow, batch_contextualized)
                batch_loss = rst_dict["loss"]
                running_loss += batch_loss
                optimizer.zero_grad()
                batch_loss.backward()
                optimizer.step()
            status.update(
                f"Fitting model. Epoch [{epoch}/{self.n_epochs}], Loss [{running_loss}]"
            )
        self.components_ = self.model.get_beta().detach().numpy()

    def fit_transform(
        self, raw_documents, y=None, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        console = Console()
        with console.status("Fitting model") as status:
            if embeddings is None:
                status.update("Encoding documents")
                embeddings = self.encoder_.encode(raw_documents)
                console.log("Documents encoded.")
            self.train_doc_embeddings = embeddings
            status.update("Extracting terms.")
            document_term_matrix = self.vectorizer.fit_transform(raw_documents)
            console.log("Term extraction done.")
            status.update("Fitting model")
            self._train_model(embeddings, document_term_matrix, status)
            console.log("Model fitting done.")
            document_topic_matrix = self.transform(
                raw_documents, embeddings=embeddings
            )
            self.top_documents = self.get_top_documents(
                raw_documents, document_topic_matrix=document_topic_matrix
            )
        return document_topic_matrix

    def transform(
        self, raw_documents, embeddings: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Infers topic importances for new documents based on a fitted model.

        Parameters
        ----------
        raw_documents: iterable of str
            Documents to fit the model on.
        embeddings: ndarray of shape (n_documents, n_dimensions), optional
            Precomputed document encodings.

        Returns
        -------
        ndarray of shape (n_dimensions, n_topics)
            Document-topic matrix.
        """
        if embeddings is None:
            embeddings = self.encoder_.encode(raw_documents)
        with torch.no_grad():
            self.model.eval()
            theta = self.model.get_theta(
                torch.as_tensor(embeddings),
                torch.as_tensor(self.train_doc_embeddings),
            )
            theta = theta.detach().cpu().numpy()
        return theta
